/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.cfa;

import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.net2plan.interfaces.networkDesign.Configuration;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.libraries.IPUtils;
import com.net2plan.libraries.SRGUtils;
import com.net2plan.libraries.WDMUtils;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This algorithm computes the number of IP links between two nodes, and the OSPF 
 * routing (with ECMP), so that all the traffic is carried, while all the links 
 * have an utilization not exceeding a given threshold ({@code maximumUtilization}). All 
 * IP links have the same given {@code fixedOSPFWeight} weight, and the 
 * same given {@code fixedIPLinkCapacity} capacity. Initially, only one IP 
 * link is established between those nodes which have a link between them in 
 * the input design. Then, we compute the worse utilization among all the 
 * links, in no failure condition, or including also any single bidirectional 
 * failure condition, if {@code singleBidirectionalFailureProvision=true}. 
 * If the worse utilization in any IP link does not exceed {@code maximumUtilization}, 
 * we are done. If not, we add two more links of {@code fixedIPLinkCapacity} 
 * (one in each direction), between the nodes of the highest utilized IP link, 
 * and iterate again.
 * 
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class CFA_OSPF_fixedWeight implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		if (netPlan.getNumberOfLayers() != 2) throw new Net2PlanException("Input design should have two layers: higher is IP, lower is WDM");
		Iterator<Long> layerIt = netPlan.getLayerIds().iterator();
		final long wdmLayerId = layerIt.next();
		final long ipLayerId = layerIt.next();

		final int N = netPlan.getNumberOfNodes();
		final int E_f = netPlan.getNumberOfLinks(wdmLayerId);
		final int D_ip = netPlan.getNumberOfDemands(ipLayerId);
		if (N == 0 || E_f == 0 || D_ip == 0) throw new Net2PlanException("This algorithm requires a physical topology (nodes and links) and demands");

		Map<Long, Pair<Long, Long>> fiberMap = netPlan.getLinkMap(wdmLayerId);
		if (!GraphUtils.isSimple(fiberMap) || !GraphUtils.isBidirectional(fiberMap)) throw new Net2PlanException("Simple and bidirectional input topologies are expected: at most one link between two nodes");

		/* Initialize some variables */
		final boolean singleBidirectionalFailureProvision = Boolean.parseBoolean(algorithmParameters.get("singleBidirectionalFailureProvision"));
		final double maximumUtilization = Double.parseDouble(algorithmParameters.get("maximumUtilization"));
		final double fixedIPLinkCapacity = Double.parseDouble(algorithmParameters.get("fixedIPLinkCapacity"));
		final double fixedOSPFWeight = Double.parseDouble(algorithmParameters.get("fixedOSPFWeight"));
		final double defaultMTTF_hours = Double.parseDouble(algorithmParameters.get("defaultMTTF_hours"));
		final double defaultMTTR_hours = Double.parseDouble(algorithmParameters.get("defaultMTTR_hours"));

		final double PRECISION_FACTOR = Double.parseDouble(Configuration.getOption("precisionFactor"));
		
		/* Initialize multilayer design (links and demands in the input design are fibers in the lower layer and demands in the upper layer, respectively) */
		netPlan.removeAllRoutingInformation(wdmLayerId);
		netPlan.removeAllRoutingInformation(ipLayerId);
		netPlan.setLayerName(wdmLayerId , "WDM");
		netPlan.setRoutingType(wdmLayerId, Constants.RoutingType.SOURCE_ROUTING);
		netPlan.setLayerName(ipLayerId , "IP");
		netPlan.setRoutingType(ipLayerId, Constants.RoutingType.HOP_BY_HOP_ROUTING);
		netPlan.setLayerDefaultId(wdmLayerId);

		/* Set an arbitrarily large number of WDM channels per fiber */
		Set<Long> fiberIds = netPlan.getLinkIds(wdmLayerId);
		for (long fiberId : fiberIds)
			WDMUtils.setFiberNumWavelengths(netPlan, fiberId, Integer.MAX_VALUE);
		
		/* Initialize IP layer with a link per each node pair with a fiber between them */
		netPlan.removeAllLinks(ipLayerId);
		for (long fiberId : fiberIds) addLightpathToFiber(netPlan, fiberId, wdmLayerId, ipLayerId, fixedIPLinkCapacity, fixedOSPFWeight);
		
		/* Configure SRGs */
		if (singleBidirectionalFailureProvision)
			SRGUtils.configureSRGs(netPlan, wdmLayerId, defaultMTTF_hours, defaultMTTR_hours, SRGUtils.SharedRiskModel.PER_BIDIRECTIONAL_LINK_BUNDLE, true);
		
		Set<Long> nodeIds = netPlan.getNodeIds();
		DoubleMatrix2D trafficMatrix = netPlan.getTrafficMatrix(ipLayerId);
		while (true)
		{
			/* Compute the failure-free routing */
			Map<Long, Double> ipLinkWeightMap = IPUtils.getLinkWeightMap(netPlan, ipLayerId);
			IPUtils.setECMPForwardingRulesFromLinkWeights(netPlan, ipLayerId, ipLinkWeightMap);
			
			/* Compute the worse carried traffic per IP link in any failure state */
			Map<Long, Double> worseCarriedTraffic_lp = new LinkedHashMap<Long, Double>(netPlan.getLinkCarriedTrafficMap(ipLayerId));
			if (singleBidirectionalFailureProvision)
			{
				long[] lpIds = netPlan.getLinkIdsVector(ipLayerId);
				Map<Long, Pair<Long, Long>> lightpathMap = netPlan.getLinkMap(ipLayerId);
				Set<Long> srgIds = netPlan.getSRGIds();
				for (long srgId : srgIds)
				{
					NetPlan netPlan_thisFailure = netPlan.copy();
					netPlan_thisFailure.setSRGDown(srgId);
					
					/* Update IGP weights according to this failure state state */
					Map<Long, Double> lpWeights_thisSRG = new LinkedHashMap<Long, Double>(ipLinkWeightMap);
					Set<Long> unavailableLightpaths = new LinkedHashSet<Long>();
					unavailableLightpaths.addAll(netPlan_thisFailure.getLinksDown(ipLayerId)); /* Lightpaths down (i.e. add/drop node is down) */
					unavailableLightpaths.addAll(netPlan_thisFailure.getLinksWithZeroCapacity(ipLayerId)); /* Lightpaths with zero capacity (i.e. some fiber is down) */
					for (long lightpathId : unavailableLightpaths) lpWeights_thisSRG.put(lightpathId, Double.MAX_VALUE);
					
					/* Compute the new routing and carried traffic per lightpath */
					final DoubleMatrix2D f_te_withoutFailingLinks = IPUtils.computeECMPRoutingTableMatrix(nodeIds, lightpathMap, lpWeights_thisSRG);
					final DoubleMatrix2D x_te = GraphUtils.convert_fte2xte(nodeIds, lightpathMap, f_te_withoutFailingLinks, trafficMatrix).getFirst();
					final double[] y_e_withoutFailingLinksVector = GraphUtils.convert_xte2ye(x_te);			

					/* Update the worse traffic per lightpath in this routing */
					for (int lp = 0; lp < lpIds.length; lp++)
					{
						long lightpathId = lpIds[lp];
						double y_e_thisLink = y_e_withoutFailingLinksVector[lp];
						if (y_e_thisLink > 0 && unavailableLightpaths.contains(lightpathId)) throw new Net2PlanException("Bad");
						worseCarriedTraffic_lp.put(lightpathId, Math.max(worseCarriedTraffic_lp.get(lightpathId), y_e_thisLink));
					}
				}
			}
			
			/* Sort the worse carried traffic map in descending order */
			worseCarriedTraffic_lp = CollectionUtils.sort(worseCarriedTraffic_lp, Constants.OrderingType.DESCENDING);

			/* If the most-subscribed lightpath is not subscribed above maximum utilization permitted, we are done */
			long mostSubscribedLightpathId = worseCarriedTraffic_lp.keySet().iterator().next();
			if (worseCarriedTraffic_lp.get(mostSubscribedLightpathId) <= maximumUtilization * fixedIPLinkCapacity + PRECISION_FACTOR) break;
			
			/* Otherwise, add a new lightpath between two nodes in both directions */
			long originNodeId = netPlan.getLinkOriginNode(ipLayerId, mostSubscribedLightpathId);
			long destinationNodeId = netPlan.getLinkDestinationNode(ipLayerId, mostSubscribedLightpathId);
			long fiberId_downstream = netPlan.getNodePairLinks(wdmLayerId, originNodeId, destinationNodeId).iterator().next();
			long fiberId_upstream = netPlan.getNodePairLinks(wdmLayerId, destinationNodeId, originNodeId).iterator().next();
			addLightpathToFiber(netPlan, fiberId_downstream, wdmLayerId, ipLayerId, fixedIPLinkCapacity, fixedOSPFWeight);
			addLightpathToFiber(netPlan, fiberId_upstream, wdmLayerId, ipLayerId, fixedIPLinkCapacity, fixedOSPFWeight);
		}
		
		/* Set the number of WDM associated to each fiber as the number of lightpaths traversing it */
		for (long fiberId : fiberIds)
			WDMUtils.setFiberNumWavelengths(netPlan, wdmLayerId, fiberId, netPlan.getLinkAssociatedRoutes(wdmLayerId, fiberId).size());
		
		WDMUtils.checkConsistency(netPlan, wdmLayerId);

		return "Ok!";
	}

	@Override
	public String getDescription()
	{
		return "This algorithm computes the number of IP links between two nodes, and the OSPF routing (with ECMP), so that all the traffic is carried, "
				+ "while all the links have an utilization not exceeding a given threshold (maximumUtilization). All IP links have the same given fixedOSPFWeight weight, "
				+ "and the same given fixedIPLinkCapacity capacity. Initially, only one IP link is established between those nodes "
				+ "which have a link between them in the input design. Then, we compute the worse utilization among all the links, in no failure condition, or including also"
				+ " any single bidirectional failure condition, if singleBidirectionalFailureProvision=true."
				+ "If the worse utilization in any IP link does not exceed maximumUtilization, we are done. If not, we add two more links of"
				+ " fixedIPLinkCapacity (one in each direction), between the nodes of the highest utilized IP link, and iterate again.";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("singleBidirectionalFailureProvision", "#boolean# true", "If true, the number of links between the nodes is computed so that if all the links between two nodes (both directions) fail simultaneously, all the traffic is carried (same OSPF weights in the surviving links) without exceeding the maximum utilization."));
		algorithmParameters.add(Triple.of("maximumUtilization", "0.8", "Maximum IP link utilization permitted in the network"));
		algorithmParameters.add(Triple.of("fixedIPLinkCapacity", "10", "All the IP links have this capacity"));
		algorithmParameters.add(Triple.of("fixedOSPFWeight", "10", "Fixed OSPF weight to assign to all the IP links"));
		algorithmParameters.add(Triple.of("defaultMTTF_hours", "8748", "Default mean time to fail (in hours)"));
		algorithmParameters.add(Triple.of("defaultMTTR_hours", "12", "Default mean time to repair (in hours)"));

		return algorithmParameters;
	}
	
	private void addLightpathToFiber(NetPlan netPlan, long fiberId, long wdmLayerId, long ipLayerId, double fixedIPLinkCapacity, double fixedOSPFWeight)
	{
		long originNodeId = netPlan.getLinkOriginNode(wdmLayerId, fiberId);
		long destinationNodeId = netPlan.getLinkDestinationNode(wdmLayerId, fiberId);
		long ipLinkId = netPlan.addLink(ipLayerId, originNodeId, destinationNodeId, fixedIPLinkCapacity, netPlan.getLinkLengthInKm(wdmLayerId, fiberId), null);
		IPUtils.setLinkWeight(netPlan, ipLayerId, ipLinkId, fixedOSPFWeight);

		/* Couple IP link with WDM demand */
		long wdmDemandId = netPlan.createLowerLayerDemandFromLink(ipLayerId, ipLinkId, wdmLayerId);

		/* Lightpath route is single hop */
		List<Long> seqFibers = new LinkedList<Long>(); seqFibers.add(fiberId);
		long wdmRouteId = netPlan.addRoute(wdmLayerId, wdmDemandId, fixedIPLinkCapacity, 1, seqFibers, null);
		int wavelengthId = netPlan.getNodePairLinks(ipLayerId, originNodeId, destinationNodeId).size() - 1;
		WDMUtils.setLightpathSeqWavelengths(netPlan, wdmLayerId, wdmRouteId, wavelengthId);
	}
}